/*
 * Copyright IBM Corp All Rights Reserved
 *
 * SPDX-License-Identifier: Apache-2.0
 */

package main

import (
	"encoding/base64"
	"fmt"

	"gmssl"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/hyperledger/fabric/protos/peer"
)

const DEFAULT_USER_ID = "1234567812345678"

// SimpleAsset implements a simple chaincode to manage an asset
type SimpleAsset struct {
}

// Init is called during chaincode instantiation to initialize any
// data. Note that chaincode upgrade also calls this function to reset
// or to migrate data.
func (t *SimpleAsset) Init(stub shim.ChaincodeStubInterface) peer.Response {
	// Get the args from the transaction proposal
	args := stub.GetStringArgs()
	if len(args) != 0 {
		return shim.Error("Incorrect arguments. Expecting a key and a value")
	}
	return shim.Success(nil)
}

// Invoke is called per transaction on the chaincode. Each transaction is
// either a 'get' or a 'set' on the asset created by Init function. The Set
// method may create a new asset by specifying a new key-value pair.
func (t *SimpleAsset) Invoke(stub shim.ChaincodeStubInterface) peer.Response {
	// Extract the function and args from the transaction proposal
	fn, args := stub.GetFunctionAndParameters()

	var result string
	var err error

	switch fn {
	case "set": {
		result, err = set(stub, args)
		}
	case "get": {
		result, err = get(stub, args)
		}
	case "verify": {
		result, err = verify(stub, args)
	}
	default: {
		shim.Error("invalid function")
	}
	}
	if err != nil {
		return shim.Error(err.Error())
	}

	// Return the result as success payload
	return shim.Success([]byte(result))
}

// Set stores the asset (both key and value) on the ledger. If the key exists,
// it will override the value with the new one
func set(stub shim.ChaincodeStubInterface, args []string) (string, error) {
	if len(args) != 2 {
		return "", fmt.Errorf("Incorrect arguments. Expecting a key and a value")
	}

	err := stub.PutState(args[0], []byte(args[1]))
	if err != nil {
		return "", fmt.Errorf("Failed to set asset: %s", args[0])
	}
	return "Success", nil
}

// Get returns the value of the specified asset key
func get(stub shim.ChaincodeStubInterface, args []string) (string, error) {
	if len(args) != 2 {
		return "", fmt.Errorf("Incorrect arguments. Expecting a user name and a secret key")
	}

	skPem, err := base64.StdEncoding.DecodeString(args[1])
	if err != nil {
		return "", err
	}

	sk, err := gmssl.NewPrivateKeyFromPEM(string(skPem), "")
	if err != nil {
		return "", fmt.Errorf("invalid secret key")
	}

	value, err := stub.GetState(args[0])
	if err != nil {
		return "", fmt.Errorf("Failed to get asset: %s with error: %s", args[0], err)
	}
	if value == nil {
		return "", fmt.Errorf("Asset not found: %s", args[0])
	}

	ciphertext, err := base64.StdEncoding.DecodeString(string(value))

	plain, err := sk.Decrypt("sm2encrypt-with-sm3", ciphertext, nil)
	if err != nil {
		return "", fmt.Errorf(err.Error())
	}
	return string(plain), nil
}

func verify(stub shim.ChaincodeStubInterface, args []string) (string, error) {
	if len(args) != 3 {
		return "", fmt.Errorf("Incorrect arguments. Expecting a message, a signature, and a public key")
	}

	pkPem, err := base64.StdEncoding.DecodeString(args[2])
	if err != nil {
		fmt.Println(err.Error())
		return "", err
	}

	pk, err := gmssl.NewPublicKeyFromPEM(string(pkPem))
	if err != nil {
                fmt.Println(err.Error())
		return "", err
	}

	sig, err := base64.StdEncoding.DecodeString(args[1])
	if err != nil {
                fmt.Println(err.Error())
		return "", err
	}

	message := args[0]

	zid, err := pk.ComputeSM2IDDigest(DEFAULT_USER_ID)
	if err != nil {
                fmt.Println(err.Error())
		return "", err
	}
	sm3ctx, err := gmssl.NewDigestContext("SM3")
	if err != nil {
                fmt.Println(err.Error())
		return "", err
	}
	sm3ctx.Reset()
	sm3ctx.Update(zid)
	sm3ctx.Update([]byte(message))
	tbs, err := sm3ctx.Final()
	if err != nil {
                fmt.Println(err.Error())
		return "", err
	}
	err = pk.Verify("sm2sign", tbs, sig, nil)
	if err == nil {
		fmt.Println("Valid Signature")
		return "Valid Signature.", nil
	} else {
		fmt.Println("Invalid Signature")
		return "Invalid signature:", nil
	}
}

// main function starts up the chaincode in the container during instantiate
func main() {
	if err := shim.Start(new(SimpleAsset)); err != nil {
		fmt.Printf("Error starting SimpleAsset chaincode: %s", err)
	}
}
